import json
import logging
import threading
from datetime import datetime, timedelta
from typing import Optional, Callable

import requests
from pydantic import BaseModel

from rapidata.service.credential_manager import CredentialManager

logger = logging.getLogger(__name__)


class TokenInfo(BaseModel):
    access_token: str
    expires_in: int
    issued_at: datetime
    token_type: str = "Bearer"

    @property
    def auth_header(self):
        return f"{self.token_type} {self.access_token}"

    @property
    def time_remaining(self):
        remaining = (
            (self.issued_at + timedelta(seconds=self.expires_in)) - datetime.now()
        ).total_seconds()
        return max(0.0, remaining)


class TokenManager:
    def __init__(
        self,
        client_id: str | None = None,
        client_secret: str | None = None,
        endpoint: str = "https://auth.rapidata.ai",
        oauth_scope: str = "openid profile email",
        cert_path: str | None = None,
        refresh_threshold: float = 0.8,
        max_sleep_time: float = 30,
    ):
        self._client_id = client_id
        self._client_secret = client_secret

        if not client_id or not client_secret:
            credential_manager = CredentialManager(
                endpoint=endpoint, cert_path=cert_path
            )
            credentials = credential_manager.get_client_credentials()
            if not credentials:
                raise ValueError("Failed to fetch client credentials")
            self._client_id = credentials.client_id
            self._client_secret = credentials.client_secret

        self._endpoint = endpoint
        self._oauth_scope = oauth_scope
        self._cert_path = cert_path
        self._refresh_threshold = refresh_threshold
        self._max_sleep_time = max_sleep_time

        self._token_lock = threading.Lock()
        self._current_token: Optional[TokenInfo] = None
        self._refresh_thread: Optional[threading.Thread] = None
        self._should_stop = threading.Event()

    def fetch_token(self):
        try:
            response = requests.post(
                f"{self._endpoint}/connect/token",
                data={
                    "grant_type": "client_credentials",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "scope": self._oauth_scope,
                },
                verify=self._cert_path,
            )

            if response.ok:
                data = response.json()
                return TokenInfo(
                    access_token=data["access_token"],
                    token_type=data["token_type"],
                    expires_in=data["expires_in"],
                    issued_at=datetime.now(),
                )

            else:
                data = response.text
                error_description = "An unknown error occurred"
                if "error_description" in data:
                    error_description = (
                        data.split("error_description")[1].split("\n")[0].strip()
                    )
                raise ValueError(f"Failed to fetch token: {error_description}")
        except requests.RequestException as e:
            raise ValueError(f"Failed to fetch token: {e}")
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse token response: {e}")
        except KeyError as e:
            raise ValueError(f"Failed to extract token from response: {e}")

    def start_token_refresh(self, token_callback: Callable[[TokenInfo], None]) -> None:
        if self._refresh_thread and self._refresh_thread.is_alive():
            logger.error("Token refresh thread is already running")
            return

        def refresh_loop():
            while not self._should_stop.is_set():
                try:
                    with self._token_lock:
                        if self._should_refresh_token(self._current_token):
                            logger.debug("Refreshing token")
                            self._current_token = self.fetch_token()
                            token_callback(self._current_token)

                    if self._current_token:
                        time_until_refresh_threshold = (
                            self._current_token.time_remaining
                            - (
                                self._current_token.expires_in
                                * (1 - self._refresh_threshold)
                            )
                        )
                        logger.debug("Time until refresh threshold: %s", time_until_refresh_threshold)
                        sleep_time = min(
                            self._max_sleep_time, time_until_refresh_threshold
                        )
                        logger.debug(
                            f"Sleeping for {sleep_time} until checking the token again"
                        )
                        self._should_stop.wait(timeout=max(1.0, sleep_time))
                    else:
                        self._should_stop.wait(timeout=self._max_sleep_time)
                except Exception as e:
                    logger.error("Failed to refresh token: %s", e)
                    self._should_stop.wait(timeout=5)

        self._should_stop.clear()
        self._refresh_thread = threading.Thread(target=refresh_loop, daemon=True)
        self._refresh_thread.start()

    def stop_token_refresh(self):
        self._should_stop.set()
        if self._refresh_thread:
            self._refresh_thread.join(timeout=1)
            self._refresh_thread = None

    def get_current_token(self) -> Optional[TokenInfo]:
        with self._token_lock:
            return self._current_token

    def _should_refresh_token(self, token: TokenInfo | None) -> bool:
        if not token:
            return True

        limit = token.expires_in * (1 - self._refresh_threshold)

        logger.debug(
            "The token was issued at %s, it expires in %s. It has %s seconds remaining and we refresh the token when it has %s seconds remaining",
            token.issued_at,
            token.expires_in,
            token.time_remaining,
            limit,
        )
        return token.time_remaining < limit

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_token_refresh()
        return False



from abc import abstractmethod
from typing import Any


class Metadata:
    """The base class for all Rapidata Metadata."""

    def __init__(self, identifier: str):
        self._identifier = identifier

    @abstractmethod
    def _to_model(self) -> Any:
        pass



from rapidata.api_client.api.campaign_api import CampaignApi
from rapidata.api_client.api.dataset_api import DatasetApi
from rapidata.api_client.api.order_api import OrderApi
from rapidata.api_client.api.pipeline_api import PipelineApi
from rapidata.api_client.api.rapid_api import RapidApi
from rapidata.api_client.api.validation_api import ValidationApi
from rapidata.api_client.api.workflow_api import WorkflowApi
from rapidata.api_client.api_client import ApiClient
from rapidata.api_client.configuration import Configuration
from rapidata.service.token_manager import TokenManager, TokenInfo

from importlib.metadata import version, PackageNotFoundError


class OpenAPIService:
    def __init__(
        self,
        client_id: str | None,
        client_secret: str | None,
        enviroment: str,
        oauth_scope: str,
        cert_path: str | None = None,
    ):
        self.enviroment = enviroment
        endpoint = f"https://api.{enviroment}"
        token_url = f"https://auth.{enviroment}"
        token_manager = TokenManager(
            client_id=client_id,
            client_secret=client_secret,
            endpoint=token_url,
            oauth_scope=oauth_scope,
            cert_path=cert_path,
        )
        client_configuration = Configuration(host=endpoint, ssl_ca_cert=cert_path)
        self.api_client = ApiClient(
            configuration=client_configuration,
            header_name="X-Client",
            header_value=f"RapidataPythonSDK/{self._get_rapidata_package_version()}",
        )

        self.api_client.configuration.api_key["bearer"] = (
            f"Bearer {token_manager.fetch_token().access_token}"
        )

        self._client_id = client_id
        self._client_secret = client_secret
        self._oauth_scope = oauth_scope
        self._token_url = f"{token_url}/connect/token"
        self._cert_path = cert_path

        token_manager.start_token_refresh(token_callback=self._set_token)

    @property
    def order_api(self) -> OrderApi:
        return OrderApi(self.api_client)

    @property
    def dataset_api(self) -> DatasetApi:
        return DatasetApi(self.api_client)

    @property
    def validation_api(self) -> ValidationApi:
        return ValidationApi(self.api_client)

    @property
    def rapid_api(self) -> RapidApi:
        return RapidApi(self.api_client)

    @property
    def campaign_api(self) -> CampaignApi:
        return CampaignApi(self.api_client)

    @property
    def pipeline_api(self) -> PipelineApi:
        return PipelineApi(self.api_client)

    @property
    def workflow_api(self) -> WorkflowApi:
        return WorkflowApi(self.api_client)

    def _set_token(self, token: TokenInfo):
        self.api_client.configuration.api_key["bearer"] = f"Bearer {token.access_token}"

    def _get_rapidata_package_version(self):
        """
        Returns the version of the currently installed rapidata package.

        Returns:
            Optional[str]: The version of the installed rapidata package,
                        or None if the package is not found
        """
        try:
            return version("rapidata")
        except PackageNotFoundError:
            return None



"""Assets Package

This package provides classes for different types of assets, including MediaAsset, TextAsset, and MultiAsset.
"""
from ._base_asset import BaseAsset
from ._media_asset import MediaAsset
from ._text_asset import TextAsset
from ._multi_asset import MultiAsset
from .data_type_enum import RapidataDataTypes



from pydantic import StrictBytes, StrictStr
from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.metadata import Metadata
from typing import Sequence
from typing import Any
from rapidata.api_client.models.add_validation_rapid_model import (
    AddValidationRapidModel,
)
from rapidata.api_client.models.add_validation_text_rapid_model import (
    AddValidationTextRapidModel,
)
from rapidata.api_client.models.add_validation_rapid_model_payload import (
    AddValidationRapidModelPayload,
)
from rapidata.api_client.models.add_validation_rapid_model_truth import (
    AddValidationRapidModelTruth,
)

from rapidata.api_client.models.datapoint_metadata_model_metadata_inner import (
    DatapointMetadataModelMetadataInner,
)


class Rapid():
    def __init__(self, asset: MediaAsset | TextAsset | MultiAsset, metadata: Sequence[Metadata], payload: Any, truth: Any, randomCorrectProbability: float, explanation: str | None):
        self.asset = asset
        self.metadata = metadata
        self.payload = payload
        self.truth = truth
        self.randomCorrectProbability = randomCorrectProbability
        self.explanation = explanation 

    def to_media_model(self, validationSetId: str) -> tuple[AddValidationRapidModel, list[StrictStr | tuple[StrictStr, StrictBytes] | StrictBytes]]:
        assets: list[MediaAsset] = [] 
        if isinstance(self.asset, MultiAsset):
            for asset in self.asset.assets:
                if isinstance(asset, MediaAsset):
                    assets.append(asset)
                else:
                    raise TypeError("The asset is a multiasset, but not all assets are MediaAssets")

        if isinstance(self.asset, TextAsset):
            raise TypeError("The asset must contain Media")

        if isinstance(self.asset, MediaAsset):
            assets = [self.asset]

        return (AddValidationRapidModel(
            validationSetId=validationSetId,
            payload=AddValidationRapidModelPayload(self.payload),
            truth=AddValidationRapidModelTruth(self.truth),
            metadata=[
              DatapointMetadataModelMetadataInner(meta._to_model())
              for meta in self.metadata
            ],
            randomCorrectProbability=self.randomCorrectProbability,
            explanation=self.explanation
        ), [asset.to_file() for asset in assets])

    def to_text_model(self, validationSetId: str) -> AddValidationTextRapidModel:
        texts: list[str] = []
        if isinstance(self.asset, MultiAsset):
            for asset in self.asset.assets:
                if isinstance(asset, TextAsset):
                    texts.append(asset.text)
                else:
                    raise TypeError("The asset is a multiasset, but not all assets are TextAssets")

        if isinstance(self.asset, MediaAsset):
            raise TypeError("The asset must contain Text")

        if isinstance(self.asset, TextAsset):
            texts = [self.asset.text]

        return AddValidationTextRapidModel(
          validationSetId=validationSetId,
          payload=AddValidationRapidModelPayload(self.payload),
          truth=AddValidationRapidModelTruth(self.truth),
          metadata=[
              DatapointMetadataModelMetadataInner(meta._to_model())
              for meta in self.metadata
          ],
          randomCorrectProbability=self.randomCorrectProbability,
          texts=texts,
          explanation=self.explanation
      )



# coding: utf-8

"""
    Rapidata.Dataset

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: v1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
import pprint
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Any, List, Optional
from rapidata.api_client.models.classification_metadata_model import ClassificationMetadataModel
from rapidata.api_client.models.count_metadata_model import CountMetadataModel
from rapidata.api_client.models.demographic_metadata_model import DemographicMetadataModel
from rapidata.api_client.models.image_dimension_metadata_model import ImageDimensionMetadataModel
from rapidata.api_client.models.location_metadata_model import LocationMetadataModel
from rapidata.api_client.models.original_filename_metadata_model import OriginalFilenameMetadataModel
from rapidata.api_client.models.prompt_metadata_model import PromptMetadataModel
from rapidata.api_client.models.text_metadata_model import TextMetadataModel
from rapidata.api_client.models.transcription_metadata_model import TranscriptionMetadataModel
from rapidata.api_client.models.translated_prompt_metadata_model import TranslatedPromptMetadataModel
from pydantic import StrictStr, Field
from typing import Union, List, Set, Optional, Dict
from typing_extensions import Literal, Self

FILEASSETMODEL1METADATAINNER_ONE_OF_SCHEMAS = ["ClassificationMetadataModel", "CountMetadataModel", "DemographicMetadataModel", "ImageDimensionMetadataModel", "LocationMetadataModel", "OriginalFilenameMetadataModel", "PromptMetadataModel", "TextMetadataModel", "TranscriptionMetadataModel", "TranslatedPromptMetadataModel"]

class FileAssetModel1MetadataInner(BaseModel):
    """
    FileAssetModel1MetadataInner
    """
    # data type: TranslatedPromptMetadataModel
    oneof_schema_1_validator: Optional[TranslatedPromptMetadataModel] = None
    # data type: ClassificationMetadataModel
    oneof_schema_2_validator: Optional[ClassificationMetadataModel] = None
    # data type: CountMetadataModel
    oneof_schema_3_validator: Optional[CountMetadataModel] = None
    # data type: DemographicMetadataModel
    oneof_schema_4_validator: Optional[DemographicMetadataModel] = None
    # data type: ImageDimensionMetadataModel
    oneof_schema_5_validator: Optional[ImageDimensionMetadataModel] = None
    # data type: LocationMetadataModel
    oneof_schema_6_validator: Optional[LocationMetadataModel] = None
    # data type: OriginalFilenameMetadataModel
    oneof_schema_7_validator: Optional[OriginalFilenameMetadataModel] = None
    # data type: PromptMetadataModel
    oneof_schema_8_validator: Optional[PromptMetadataModel] = None
    # data type: TextMetadataModel
    oneof_schema_9_validator: Optional[TextMetadataModel] = None
    # data type: TranscriptionMetadataModel
    oneof_schema_10_validator: Optional[TranscriptionMetadataModel] = None
    actual_instance: Optional[Union[ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel]] = None
    one_of_schemas: Set[str] = { "ClassificationMetadataModel", "CountMetadataModel", "DemographicMetadataModel", "ImageDimensionMetadataModel", "LocationMetadataModel", "OriginalFilenameMetadataModel", "PromptMetadataModel", "TextMetadataModel", "TranscriptionMetadataModel", "TranslatedPromptMetadataModel" }

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )


    discriminator_value_class_map: Dict[str, str] = {
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_oneof(cls, v):
        instance = FileAssetModel1MetadataInner.model_construct()
        error_messages = []
        match = 0
        # validate data type: TranslatedPromptMetadataModel
        if not isinstance(v, TranslatedPromptMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `TranslatedPromptMetadataModel`")
        else:
            match += 1
        # validate data type: ClassificationMetadataModel
        if not isinstance(v, ClassificationMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `ClassificationMetadataModel`")
        else:
            match += 1
        # validate data type: CountMetadataModel
        if not isinstance(v, CountMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `CountMetadataModel`")
        else:
            match += 1
        # validate data type: DemographicMetadataModel
        if not isinstance(v, DemographicMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `DemographicMetadataModel`")
        else:
            match += 1
        # validate data type: ImageDimensionMetadataModel
        if not isinstance(v, ImageDimensionMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `ImageDimensionMetadataModel`")
        else:
            match += 1
        # validate data type: LocationMetadataModel
        if not isinstance(v, LocationMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `LocationMetadataModel`")
        else:
            match += 1
        # validate data type: OriginalFilenameMetadataModel
        if not isinstance(v, OriginalFilenameMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `OriginalFilenameMetadataModel`")
        else:
            match += 1
        # validate data type: PromptMetadataModel
        if not isinstance(v, PromptMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `PromptMetadataModel`")
        else:
            match += 1
        # validate data type: TextMetadataModel
        if not isinstance(v, TextMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `TextMetadataModel`")
        else:
            match += 1
        # validate data type: TranscriptionMetadataModel
        if not isinstance(v, TranscriptionMetadataModel):
            error_messages.append(f"Error! Input type `{type(v)}` is not `TranscriptionMetadataModel`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when setting `actual_instance` in FileAssetModel1MetadataInner with oneOf schemas: ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when setting `actual_instance` in FileAssetModel1MetadataInner with oneOf schemas: ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel. Details: " + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Union[str, Dict[str, Any]]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into TranslatedPromptMetadataModel
        try:
            instance.actual_instance = TranslatedPromptMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into ClassificationMetadataModel
        try:
            instance.actual_instance = ClassificationMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into CountMetadataModel
        try:
            instance.actual_instance = CountMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into DemographicMetadataModel
        try:
            instance.actual_instance = DemographicMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into ImageDimensionMetadataModel
        try:
            instance.actual_instance = ImageDimensionMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into LocationMetadataModel
        try:
            instance.actual_instance = LocationMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into OriginalFilenameMetadataModel
        try:
            instance.actual_instance = OriginalFilenameMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into PromptMetadataModel
        try:
            instance.actual_instance = PromptMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into TextMetadataModel
        try:
            instance.actual_instance = TextMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into TranscriptionMetadataModel
        try:
            instance.actual_instance = TranscriptionMetadataModel.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when deserializing the JSON string into FileAssetModel1MetadataInner with oneOf schemas: ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when deserializing the JSON string into FileAssetModel1MetadataInner with oneOf schemas: ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel. Details: " + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> Optional[Union[Dict[str, Any], ClassificationMetadataModel, CountMetadataModel, DemographicMetadataModel, ImageDimensionMetadataModel, LocationMetadataModel, OriginalFilenameMetadataModel, PromptMetadataModel, TextMetadataModel, TranscriptionMetadataModel, TranslatedPromptMetadataModel]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())





from rapidata.rapidata_client.validation.rapidata_validation_set import RapidataValidationSet
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.assets.data_type_enum import RapidataDataTypes
from rapidata.rapidata_client.validation.rapids.rapids_manager import RapidsManager
from rapidata.rapidata_client.validation.rapids.rapids import Rapid
from rapidata.rapidata_client.metadata import PromptMetadata

from rapidata.api_client.models.page_info import PageInfo
from rapidata.api_client.models.root_filter import RootFilter
from rapidata.api_client.models.filter import Filter
from rapidata.api_client.models.sort_criterion import SortCriterion
from rapidata.api_client.exceptions import BadRequestException
from urllib3._collections import HTTPHeaderDict

from rapidata.rapidata_client.validation.rapids.box import Box

from rapidata.api_client.models.query_validation_set_model import QueryValidationSetModel


class ValidationSetManager:
    """
    Responsible for everything related to validation sets. From creation to retrieval.

    Attributes:
        rapid (RapidsManager): The RapidsManager instance.
    """
    def __init__(self, openapi_service: OpenAPIService) -> None:
        self.__openapi_service = openapi_service
        self.rapid = RapidsManager()

    def create_classification_set(self,
        name: str,
        instruction: str,
        answer_options: list[str],
        datapoints: list[str],
        truths: list[list[str]],
        data_type: str = RapidataDataTypes.MEDIA,
        contexts: list[str] | None = None,
        print_confirmation: bool = True,
        explanations: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a classification validation set.
        
        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction by which the labeler will answer.
            answer_options (list[str]): The options to choose from when answering.
            datapoints (list[str]): The datapoints that will be used for validation.
            truths (list[list[str]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    options: ["yes", "no", "maybe"]
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [["yes"], ["no", "maybe"]] -> first datapoint correct answer is "yes", second datapoint is "no" or "maybe"
            data_type (str, optional): The type of data. Defaults to RapidataDataTypes.MEDIA. Other option: RapidataDataTypes.TEXT ("text").
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and answer options. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanations (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, (list, tuple)) for truth in truths]):
            raise ValueError("Truths must be a list of lists or tuples")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")

        if(explanations and len(explanations) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
       
        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.classification_rapid(
                    instruction=instruction,
                    answer_options=answer_options,
                    datapoint=datapoints[i],
                    truths=truths[i],
                    data_type=data_type,
                    metadata=[PromptMetadata(contexts[i])] if contexts else [],
                    explanation=explanations[i] if explanations != None else None
                )
            )

        return self._submit(name=name, data_type=data_type, rapids=rapids, print_confirmation=print_confirmation)

    def create_compare_set(self,
        name: str,
        instruction: str,
        datapoints: list[list[str]],
        truths: list[str],
        data_type: str = RapidataDataTypes.MEDIA,
        contexts: list[str] | None = None,
        print_confirmation: bool = True,
        explanation: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a comparison validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to compare against.
            truths (list[str]): The truths for each comparison. List is for each comparison.\n
                example:
                    instruction: "Which image has a cat?"
                    datapoints = [["image1.jpg", "image2.jpg"], ["image3.jpg", "image4.jpg"]]
                    truths: ["image1.jpg", "image4.jpg"] -> first comparison image1.jpg has a cat, second comparison image4.jpg has a cat
            datapoints (list[list[str]]): The compare datapoints to create the validation set with. 
                Outer list is for each comparison, inner list the two images/texts that will be compared.
            data_type (str, optional): The type of data. Defaults to RapidataDataTypes.MEDIA. Other option: RapidataDataTypes.TEXT ("text").
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and truth. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanation (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, str) for truth in truths]):
            raise ValueError("Truths must be a list of strings")

        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
 
        if(explanation and len(explanation) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
              
        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.compare_rapid(
                    instruction=instruction,
                    truth=truths[i],
                    datapoint=datapoints[i],
                    data_type=data_type,
                    metadata=[PromptMetadata(contexts[i])] if contexts else [],
                    explanation=explanation[i] if explanation != None else None
                )
            )
 
        return self._submit(name=name, data_type=data_type, rapids=rapids, print_confirmation=print_confirmation)
  
    def create_select_words_set(self,
        name: str,
        instruction: str,
        truths: list[list[int]],
        datapoints: list[str],
        sentences: list[str],
        required_precision: float = 1.0,
        required_completeness: float = 1.0,
        print_confirmation: bool = True,
        explanation: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a select words validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[int]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    sentences: ["this example 1", "this example 2"]
                    truths: [[0, 1], [2]] -> first datapoint correct words are "this" and "example", second datapoint is "2"
            datapoints (list[str]): The datapoints that will be used for validation.
            sentences (list[str]): The sentences that will be used for validation. The sentece will be split up by spaces to be selected by the labeler.
                Must be the same length as datapoints.
            required_precision (float, optional): The required precision for the labeler to get the rapid correct (minimum ratio of the words selected that need to be correct). Defaults to 1.0 (no wrong word can be selected).
            required_completeness (float, optional): The required completeness for the labeler to get the rapid correct (miminum ratio of total correct words selected). Defaults to 1.0 (all correct words need to be selected).
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanation (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
            """
        
        if not all([isinstance(truth, (list, tuple)) for truth in truths]):
            raise ValueError("Truths must be a list of lists or tuples")

        if len(datapoints) != len(truths) or len(datapoints) != len(sentences):
            raise ValueError("The number of datapoints, truths, and sentences must be equal")
 
        if(explanation and len(explanation) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
              
        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.select_words_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    sentence=sentences[i],
                    required_precision=required_precision,
                    required_completeness=required_completeness,
                    explanation=explanation[i] if explanation != None else None
                )
            )

        return self._submit(name=name, data_type=RapidataDataTypes.MEDIA, rapids=rapids, print_confirmation=print_confirmation)

    def create_locate_set(self,
        name: str,
        instruction: str,
        truths: list[list[Box]],
        datapoints: list[str],
        contexts: list[str] | None = None,
        print_confirmation: bool = True,
        explanation: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a locate validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[Box]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [[Box(0, 0, 100, 100)], [Box(50, 50, 150, 150)]] -> first datapoint the object is in the top left corner, second datapoint the object is in the center
            datapoints (list[str]): The datapoints that will be used for validation.
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanation (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, (list, tuple)) for truth in truths]):
            raise ValueError("Truths must be a list of lists or tuples")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
 
        if(explanation and len(explanation) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
              
        rapids = []
        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.locate_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    metadata=[PromptMetadata(contexts[i])] if contexts else [],
                    explanation=explanation[i] if explanation != None else None

                )
            )
        
        return self._submit(name=name, data_type=RapidataDataTypes.MEDIA, rapids=rapids, print_confirmation=print_confirmation)
    
    def create_draw_set(self,
        name: str,
        instruction: str,
        truths: list[list[Box]],
        datapoints: list[str],
        contexts: list[str] | None = None,
        print_confirmation: bool = True,
        explanation: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a draw validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[Box]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [[Box(0, 0, 100, 100)], [Box(50, 50, 150, 150)]] -> first datapoint the object is in the top left corner, second datapoint the object is in the center
            datapoints (list[str]): The datapoints that will be used for validation.
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanation (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, (list, tuple)) for truth in truths]):
            raise ValueError("Truths must be a list of lists or tuples")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
 
        if(explanation and len(explanation) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
              
        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.draw_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    metadata=[PromptMetadata(contexts[i])] if contexts else [],
                    explanation=explanation[i] if explanation != None else None

                )
            )

        return self._submit(name=name, data_type=RapidataDataTypes.MEDIA, rapids=rapids, print_confirmation=print_confirmation)

    def create_timestamp_set(self,
        name: str,
        instruction: str,
        truths: list[list[tuple[int, int]]],
        datapoints: list[str],
        contexts: list[str] | None = None,
        print_confirmation: bool = True,
        explanation: list[str | None] | None = None,
    ) -> RapidataValidationSet:
        """Create a timestamp validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[tuple[int, int]]]): The truths for each datapoint defined as start and endpoint based on miliseconds. 
            Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [[(0, 10)], [(20, 30)]] -> first datapoint the correct interval is from 0 to 10, second datapoint the correct interval is from 20 to 30
            datapoints (list[str]): The datapoints that will be used for validation.
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            explanation (list[str | None], optional): The explanations for each datapoint. Will be given to the annotators in case the answer is wrong. Defaults to None.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, (list, tuple)) for truth in truths]):
            raise ValueError("Truths must be a list of lists or tuples")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
 
        if(explanation and len(explanation) != len(datapoints)):
            raise ValueError("The numeber of reasons and datapoints must be equal, the index must align, but can be padded with None")
              

        rapids: list[Rapid] = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.timestamp_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    metadata=[PromptMetadata(contexts[i])] if contexts else [],
                    explanation=explanation[i] if explanation != None else None
                )
            )

        return self._submit(name=name, data_type=RapidataDataTypes.MEDIA, rapids=rapids, print_confirmation=print_confirmation)
    
    def create_mixed_set(self,
        name: str,
        data_type: str,
        rapids: list[Rapid],
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a validation set with a list of rapids.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            rapids (list[Rapid]): The list of rapids to add to the validation set.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """

        return self._submit(name, data_type, rapids, print_confirmation)
    
    def get_validation_set_by_id(self, validation_set_id: str) -> RapidataValidationSet:
        """Get a validation set by ID.

        Args:
            validation_set_id (str): The ID of the validation set.

        Returns:
            RapidataValidationSet: The ValidationSet instance.
        """
        try:
            validation_set = self.__openapi_service.validation_api.validation_get_by_id_get(id=validation_set_id)
        except Exception:
            raise ValueError(f"ValidationSet with ID {validation_set_id} not found.")
        
        return RapidataValidationSet(validation_set_id, validation_set.name)

    def _submit(self, name: str, data_type: str, rapids: list[Rapid], print_confirmation: bool) -> RapidataValidationSet:
        validation_set_id = (
            self.__openapi_service.validation_api.validation_create_validation_set_post(
                name=name
            )
        ).validation_set_id

        if validation_set_id is None:
            raise ValueError("Failed to create validation set")

        if print_confirmation:
            print(f"Validation set '{name}' created with ID {validation_set_id}")

        for rapid in rapids:
            if data_type == RapidataDataTypes.TEXT:
                self.__openapi_service.validation_api.validation_add_validation_text_rapid_post(
                    add_validation_text_rapid_model=rapid.to_text_model(validation_set_id)
                )
            else:
                model = rapid.to_media_model(validation_set_id)
                self.__openapi_service.validation_api.validation_add_validation_rapid_post(
                   model=model[0], files=model[1]
                )

        return RapidataValidationSet(
            name=name,
            validation_set_id=validation_set_id,
        )


    def find_validation_sets(self, name: str = "", amount: int = 1) -> list[RapidataValidationSet]:
        """Find validation sets by name.

        Args:
            name (str, optional): The name to search for. Defaults to "" to match with any set.
            amount (int, optional): The amount of validation sets to return. Defaults to 1.

        Returns:
            list[RapidataValidationSet]: The list of validation sets.
        """
        try:
            validation_page_result = self.__openapi_service.validation_api.validation_query_validation_sets_get(QueryValidationSetModel(
                pageInfo=PageInfo(index=1, size=amount),
                filter=RootFilter(filters=[Filter(field="Name", operator="Contains", value=name)]),
                sortCriteria=[SortCriterion(direction="Desc", propertyName="CreatedAt")]
                ))

        except BadRequestException as e:
            raise ValueError(f"Error occured during request. \nError: {e.body} \nTraceid: {e.headers.get('X-Trace-Id') if isinstance(e.headers, HTTPHeaderDict) else 'Unknown'}")

        except Exception as e:
            raise ValueError(f"Unknown error occured: {e}")

        validation_sets = [self.get_validation_set_by_id(validation_set.id) for validation_set in validation_page_result.items]
        return validation_sets




from typing import Any
from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.campaign_user_filter_model import (
    CampaignUserFilterModel,
)


class CampaignFilter(RapidataFilter):
    """CampaignFilter Class

    Can be used to filter who to target based on campaign IDs.

    This filter can only be used when directly in contact with Rapidata.
    
    Args:
        campaign_ids (list[str]): List of campaign IDs to filter by.
    """

    def __init__(self, campaign_ids: list[str]):
        self.campaign_ids = campaign_ids

    def _to_model(self):
        return CampaignUserFilterModel(
            _t="CampaignFilter",
            campaignIds=self.campaign_ids,
        )



import json
import os
import time
import urllib.parse
import webbrowser
from datetime import datetime, timezone
from pathlib import Path
from socket import gethostname
from typing import Dict, List, Optional, Tuple

import requests
from colorama import Fore
from pydantic import BaseModel


class ClientCredential(BaseModel):
    display_name: str
    client_id: str
    client_secret: str
    endpoint: str
    created_at: datetime
    last_used: datetime

    @classmethod
    def from_dict(cls, data: Dict):
        return cls(
            display_name=data["display_name"],
            client_id=data["client_id"],
            client_secret=data["client_secret"],
            endpoint=data["endpoint"],
            created_at=datetime.fromisoformat(data["created_at"]),
            last_used=datetime.fromisoformat(data["last_used"]),
        )

    def to_dict(self) -> Dict:
        return {
            "display_name": self.display_name,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "endpoint": self.endpoint,
            "created_at": self.created_at.isoformat(),
            "last_used": self.last_used.isoformat(),
        }

    def get_display_string(self):
        return f"{self.display_name} - Client ID: {self.client_id} (Created: {self.created_at.strftime('%Y-%m-%d %H:%M:%S')})"


class BridgeToken(BaseModel):
    read_key: str
    write_key: str


class CredentialManager:
    def __init__(
        self,
        endpoint: str,
        cert_path: str | None = None,
        poll_timeout: int = 300,
        poll_interval: int = 1,
    ):
        self.endpoint = endpoint
        self.cert_path = cert_path
        self.poll_timeout = poll_timeout
        self.poll_interval = poll_interval

        self.config_dir = Path.home() / ".config" / "rapidata"
        self.config_path = self.config_dir / "credentials.json"

        # Ensure config directory exists
        self.config_dir.mkdir(parents=True, exist_ok=True)

    def _read_credentials(self) -> Dict[str, List[ClientCredential]]:
        """Read all stored credentials from the config file."""
        if not self.config_path.exists():
            return {}

        try:
            with open(self.config_path, "r") as f:
                data = json.load(f)
                return {
                    env: [ClientCredential.from_dict(cred) for cred in creds]
                    for env, creds in data.items()
                }
        except json.JSONDecodeError:
            return {}

    def _write_credentials(
        self, credentials: Dict[str, List[ClientCredential]]
    ) -> None:
        data = {
            env: [cred.to_dict() for cred in creds]
            for env, creds in credentials.items()
        }

        with open(self.config_path, "w") as f:
            json.dump(data, f, indent=2)

        # Ensure file is only readable by the user
        os.chmod(self.config_path, 0o600)

    def _store_credential(self, credential: ClientCredential) -> None:
        credentials = self._read_credentials()

        if credential.endpoint not in credentials:
            credentials[credential.endpoint] = []

        credentials[credential.endpoint].append(credential)
        self._write_credentials(credentials)

    @staticmethod
    def _select_credential(
        credentials: List[ClientCredential],
    ) -> Optional[ClientCredential]:
        if not credentials:
            return None

        if len(credentials) == 1:
            return credentials[0]

        return max(credentials, key=lambda c: c.last_used)

    def get_client_credentials(self) -> Optional[ClientCredential]:
        """Gets stored client credentials or create new ones via browser auth."""
        credentials = self._read_credentials()
        env_credentials = credentials.get(self.endpoint, [])

        if env_credentials:
            credential = self._select_credential(env_credentials)
            if credential:
                credential.last_used = datetime.now(timezone.utc)
                self._write_credentials(credentials)
                return credential

        return self._create_new_credentials()

    def _get_bridge_tokens(self) -> Optional[BridgeToken]:
        """Get bridge tokens from the identity endpoint."""
        try:
            bridge_endpoint = (
                f"{self.endpoint}/Identity/CreateBridgeToken?clientId=rapidata-cli"
            )
            response = requests.post(bridge_endpoint, verify=self.cert_path)
            if not response.ok:
                print(f"Failed to get bridge tokens: {response.status_code}")
                return None

            data = response.json()
            return BridgeToken(read_key=data["readKey"], write_key=data["writeKey"])
        except requests.RequestException as e:
            print(f"Failed to get bridge tokens: {e}")
            return None

    def _poll_read_key(self, read_key: str) -> Optional[str]:
        """Poll the read key endpoint until we get an access token."""
        read_endpoint = f"{self.endpoint}/Identity/ReadBridgeToken"
        start_time = time.time()

        while time.time() - start_time < self.poll_timeout:
            try:
                response = requests.get(
                    read_endpoint, params={"readKey": read_key}, verify=self.cert_path
                )

                if response.status_code == 200:
                    return response.json().get("accessToken")
                elif response.status_code == 202:
                    # Still processing
                    time.sleep(self.poll_interval)
                    continue
                else:
                    # Error occurred
                    print(f"Error polling read key: {response.status_code}")
                    return None

            except requests.RequestException as e:
                print(f"Error polling read key: {e}")
                return None

        print("Polling timed out")
        return None

    def _create_client(self, access_token: str) -> Optional[Tuple[str, str, str]]:
        """Create a new client using the access token."""
        try:
            # set the display name to the hostname
            display_name = f"{gethostname()} - Python API Client"
            response = requests.post(
                f"{self.endpoint}/Client",
                headers={
                    "Authorization": f"Bearer {access_token}",
                    "Content-Type": "application/json",
                    "Accept": "*/*",
                },
                json={"displayName": display_name},
                verify=self.cert_path,
            )
            response.raise_for_status()
            data = response.json()
            return data.get("clientId"), data.get("clientSecret"), display_name
        except requests.RequestException as e:
            print(f"Failed to create client: {e}")
            return None

    def _create_new_credentials(self) -> Optional[ClientCredential]:
        bridge_endpoint = self._get_bridge_tokens()
        if not bridge_endpoint:
            return None

        auth_url = f"{self.endpoint}/connect/authorize/external?clientId=rapidata-cli&scope=openid profile email roles&writeKey={bridge_endpoint.write_key}"
        could_open_browser = webbrowser.open(auth_url)

        if not could_open_browser:
            encoded_url = urllib.parse.quote(auth_url, safe="%/:=&?~#+!$,;'@()*[]")
            print(
                Fore.RED
                + f'Please open the following URL in your browser to log in: "{encoded_url}"'
                + Fore.RESET
            )

        access_token = self._poll_read_key(bridge_endpoint.read_key)
        if not access_token:
            return None

        client_state = self._create_client(access_token)

        if not client_state:
            raise ValueError("Failed to create client")

        client_id, client_secret, display_name = client_state

        credential = ClientCredential(
            client_id=client_id,
            client_secret=client_secret,
            display_name=display_name,
            endpoint=self.endpoint,
            created_at=datetime.now(timezone.utc),
            last_used=datetime.now(timezone.utc),
        )

        self._store_credential(credential)
        return credential



from .models import TranslationBehaviourOptions
from .alert_on_fast_response import AlertOnFastResponse
from .translation_behaviour import TranslationBehaviour
from .free_text_minimum_characters import FreeTextMinimumCharacters
from .no_shuffle import NoShuffle
from .play_video_until_the_end import PlayVideoUntilTheEnd
from .custom_setting import CustomSetting
from .rapidata_settings import RapidataSettings
from ._rapidata_setting import RapidataSetting



class RapidataValidationSet:
    """A class for interacting with a Rapidata validation set.

    Represents a set of all the validation tasks that can be added to an order.

    When added to an order, the tasks will be selected randomly from the set.

    Attributes:
        id (str): The ID of the validation set.
        name (str): The name of the validation set.
    """

    def __init__(self, validation_set_id, name: str):
        self.id = validation_set_id
        self.name = name

    def __str__(self):
        return f"name: '{self.name}' id: {self.id}"

    def __repr__(self):
        return f"name: '{self.name}' id: {self.id}"



from rapidata.api_client.models.box_shape import BoxShape

class Box:
    """
    Used in the Locate and Draw Validation sets. All coordinates are in pixels.
    
    Args:
        x_min (float): The minimum x value of the box.
        y_min (float): The minimum y value of the box.
        x_max (float): The maximum x value of the box.
        y_max (float): The maximum y value of the box.
    """
    def __init__(self, x_min: float, y_min: float, x_max: float, y_max: float):
        self.x_min = x_min
        self.y_min = y_min
        self.x_max = x_max
        self.y_max = y_max



from rapidata.api_client.models.age_group import AgeGroup as BackendAgeGroupModel
from enum import Enum

class AgeGroup(Enum):
    """AgeGroup Enum
    
    Represents the age group of a user. Used to filter who to target based on age groups.
    
    Attributes:
        UNDER_18 (AgeGroup): Represents the age group of users under 18.
        BETWEEN_18_29 (AgeGroup): Represents the age group of users between 18 and 29.
        BETWEEN_30_39 (AgeGroup): Represents the age group of users between 30 and 39.
        BETWEEN_40_49 (AgeGroup): Represents the age group of users between 40 and 49.
        BETWEEN_50_64 (AgeGroup): Represents the age group of users between 50 and 64.
        OVER_65 (AgeGroup): Represents the age group of users over 65."""
    
    UNDER_18 = BackendAgeGroupModel.ENUM_0_MINUS_17
    BETWEEN_18_29 = BackendAgeGroupModel.ENUM_18_MINUS_29
    BETWEEN_30_39 = BackendAgeGroupModel.ENUM_30_MINUS_39
    BETWEEN_40_49 = BackendAgeGroupModel.ENUM_40_MINUS_49
    BETWEEN_50_64 = BackendAgeGroupModel.ENUM_50_MINUS_64
    OVER_65 = BackendAgeGroupModel.ENUM_65_PLUS

    def _to_backend_model(self) -> BackendAgeGroupModel:
        return BackendAgeGroupModel(self.value)




from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.simple_workflow_model_blueprint import SimpleWorkflowModelBlueprint
from rapidata.api_client.models.scrub_rapid_blueprint import ScrubRapidBlueprint
from rapidata.rapidata_client.workflow._base_workflow import Workflow


class TimestampWorkflow(Workflow):
    """
    A workflow for timestamp tasks.

    This class represents a timestamp workflow 
    where audio or video content receives timestamps based on the instruction.

    Attributes:
        _instruction (str): The instruction for the timestamp task.

    Args:
        instruction (str): The instruction to be provided for the timestamp task.
    """

    def __init__(self, instruction: str):
        super().__init__(type="SimpleWorkflowConfig")
        self._instruction = instruction

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = ScrubRapidBlueprint(
            _t="ScrubBlueprint",
            target=self._instruction
        )

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint)
        )
